import ast
import logging
from typing import Dict, Set, Callable

from deply.models.code_element import CodeElement
from deply.models.dependency import Dependency
from deply.utils.ast_utils import set_ast_parents
from deply.utils.dependency_visitor import DependencyVisitor


class CodeAnalyzer:
    def __init__(
            self,
            code_elements: Set[CodeElement],
            dependency_handler: Callable[[Dependency], None],
    ):
        self.code_elements = code_elements
        self.dependency_handler = dependency_handler
        self.dependency_types = [
            'import',
            'import_from',
            'function_call',
            'class_inheritance',
            'decorator',
            'type_annotation',
            'exception_handling',
            'metaclass',
            'attribute_access',
            'name_load',
        ]
        logging.debug(f"Initialized CodeAnalyzer with {len(self.code_elements)} code elements.")

    def analyze(self) -> None:
        logging.debug("Starting analysis of code elements.")
        name_to_elements = self._build_name_to_element_map()
        logging.debug(f"Name to elements map built with {len(name_to_elements)} names.")

        file_to_elements: Dict[str, Set[CodeElement]] = {}
        for code_element in self.code_elements:
            file_to_elements.setdefault(code_element.file, set()).add(code_element)

        for file_path, elements_in_file in file_to_elements.items():
            logging.debug(f"Analyzing file: {file_path} with {len(elements_in_file)} code elements")
            self._extract_dependencies_from_file(file_path, elements_in_file, name_to_elements)
        logging.debug("Completed analysis of code elements.")

    def _build_name_to_element_map(self) -> Dict[str, Set[CodeElement]]:
        logging.debug("Building name to element map.")
        name_to_element = {}
        for elem in self.code_elements:
            name_to_element.setdefault(elem.name, set()).add(elem)
        logging.debug(f"Name to element map contains {len(name_to_element)} entries.")
        return name_to_element

    def _extract_dependencies_from_file(
            self,
            file_path: str,
            code_elements_in_file: Set[CodeElement],
            name_to_elements: Dict[str, Set[CodeElement]]
    ) -> None:
        logging.debug(f"Extracting dependencies from file: {file_path}")
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                source_code = f.read()
            logging.debug(f"File {file_path} read successfully.")
            tree = ast.parse(source_code, filename=str(file_path))
            set_ast_parents(tree)
            logging.debug(f"AST parsing completed for {file_path}.")
        except (SyntaxError, FileNotFoundError, UnicodeDecodeError) as e:
            logging.warning(f"Failed to parse {file_path}: {e}")
            return

        elements_in_file_by_name = {elem.name: elem for elem in code_elements_in_file}

        visitor = DependencyVisitor(
            code_elements_in_file=elements_in_file_by_name,
            dependency_types=self.dependency_types,
            dependency_handler=self.dependency_handler,
            name_to_elements=name_to_elements,
        )
        logging.debug(f"Starting AST traversal for file: {file_path}")
        visitor.visit(tree)
        logging.debug(f"Completed AST traversal for file: {file_path}")



import sys
import types
import unittest
from typing import List

from deply.collectors.base_collector import BaseCollector
from deply.collectors.collector_factory import CollectorFactory


# Dummy custom collector implementing BaseCollector interface
class DummyCustomCollector(BaseCollector):
    def __init__(self, config: dict):
        self.config = config

    def match_in_file(self, file_ast, file_path) -> set:
        return set()


class TestCollectorFactoryCustom(unittest.TestCase):
    def setUp(self):
        # Inject dummy module into sys.modules so that it can be imported
        module_name = "dummy_custom_collectors"
        self.dummy_module = types.ModuleType(module_name)
        setattr(self.dummy_module, "DummyCustomCollector", DummyCustomCollector)
        sys.modules[module_name] = self.dummy_module

    def tearDown(self):
        # Clean up the injected module
        if "dummy_custom_collectors" in sys.modules:
            del sys.modules["dummy_custom_collectors"]

    def test_custom_collector_creation(self):
        config = {
            "type": "custom",
            "class": "dummy_custom_collectors.DummyCustomCollector",
            "params": {"foo": "bar"}
        }
        paths: List[str] = ["dummy_path"]
        exclude_files: List[str] = ["dummy_pattern"]
        collector = CollectorFactory.create(config, paths, exclude_files)
        # Assert that we got an instance of our dummy collector
        self.assertIsInstance(collector, DummyCustomCollector)
        # Expected config passed to the custom collector
        expected_config = {
            "params": {"foo": "bar"},
            "paths": paths,
            "exclude_files": exclude_files,
        }
        self.assertEqual(collector.config, expected_config)

    def test_custom_collector_missing_class_field(self):
        config = {
            "type": "custom",
            "params": {"foo": "bar"}
        }
        with self.assertRaises(ValueError) as context:
            CollectorFactory.create(config, [], [])
        self.assertIn("Custom collector requires 'class' field", str(context.exception))

    def test_custom_collector_invalid_class_path(self):
        config = {
            "type": "custom",
            "class": "invalidpath",
            "params": {"foo": "bar"}
        }
        with self.assertRaises(ValueError) as context:
            CollectorFactory.create(config, [], [])
        self.assertIn("Invalid class path format", str(context.exception))

    def test_custom_collector_module_not_found(self):
        config = {
            "type": "custom",
            "class": "nonexistent_module.DummyCustomCollector",
            "params": {"foo": "bar"}
        }
        with self.assertRaises(ValueError) as context:
            CollectorFactory.create(config, [], [])
        self.assertIn("Failed to import module", str(context.exception))

    def test_custom_collector_class_not_found(self):
        config = {
            "type": "custom",
            "class": "dummy_custom_collectors.NonExistentClass",
            "params": {"foo": "bar"}
        }
        with self.assertRaises(ValueError) as context:
            CollectorFactory.create(config, [], [])
        self.assertIn("Class 'NonExistentClass' not found", str(context.exception))


if __name__ == "__main__":
    unittest.main()



import ast
import logging
from typing import Dict, List, Callable, Set

from deply.models.code_element import CodeElement
from deply.models.dependency import Dependency


class DependencyVisitor(ast.NodeVisitor):
    def __init__(
            self,
            code_elements_in_file: Dict[str, CodeElement],
            dependency_types: List[str],
            dependency_handler: Callable[[Dependency], None],
            name_to_elements: Dict[str, Set[CodeElement]],
    ):
        self.code_elements_in_file = code_elements_in_file
        self.dependency_types = dependency_types
        self.dependency_handler = dependency_handler
        self.name_to_elements = name_to_elements
        self.current_code_element = None
        logging.debug(f"DependencyVisitor created for file with {len(code_elements_in_file)} code elements")

    def visit_FunctionDef(self, node):
        full_name = self._get_definition_full_name(node)
        self.current_code_element = self.code_elements_in_file.get(full_name)
        if 'decorator' in self.dependency_types and self.current_code_element:
            self._process_decorators(node)
        if 'type_annotation' in self.dependency_types and self.current_code_element:
            if node.returns:
                self._process_annotation(node.returns)
            for arg in node.args.args + node.args.kwonlyargs:
                if arg.annotation:
                    self._process_annotation(arg.annotation)
        self.generic_visit(node)
        self.current_code_element = None

    def visit_ClassDef(self, node):
        full_name = self._get_definition_full_name(node)
        self.current_code_element = self.code_elements_in_file.get(full_name)
        if 'class_inheritance' in self.dependency_types and self.current_code_element:
            for base in node.bases:
                base_name = self._get_full_name(base)
                dep_elements = self.name_to_elements.get(base_name, set())
                for dep_element in dep_elements:
                    dependency = Dependency(
                        code_element=self.current_code_element,
                        depends_on_code_element=dep_element,
                        dependency_type='class_inheritance',
                        line=base.lineno,
                        column=base.col_offset
                    )
                    self.dependency_handler(dependency)
        if 'decorator' in self.dependency_types and self.current_code_element:
            self._process_decorators(node)
        if 'metaclass' in self.dependency_types and self.current_code_element:
            for keyword in node.keywords:
                if keyword.arg == 'metaclass':
                    metaclass_name = self._get_full_name(keyword.value)
                    dep_elements = self.name_to_elements.get(metaclass_name, set())
                    for dep_element in dep_elements:
                        dependency = Dependency(
                            code_element=self.current_code_element,
                            depends_on_code_element=dep_element,
                            dependency_type='metaclass',
                            line=keyword.value.lineno,
                            column=keyword.value.col_offset
                        )
                        self.dependency_handler(dependency)
        self.generic_visit(node)
        self.current_code_element = None

    def visit_Call(self, node):
        if 'function_call' in self.dependency_types and self.current_code_element:
            if isinstance(node.func, ast.Name):
                name = node.func.id
                dep_elements = self.name_to_elements.get(name, set())
                for dep_element in dep_elements:
                    dependency = Dependency(
                        code_element=self.current_code_element,
                        depends_on_code_element=dep_element,
                        dependency_type='function_call',
                        line=node.lineno,
                        column=node.col_offset
                    )
                    self.dependency_handler(dependency)
            elif isinstance(node.func, ast.Attribute):
                full_name = self._get_full_name(node.func)
                dep_elements = self.name_to_elements.get(full_name, set())
                for dep_element in dep_elements:
                    dependency = Dependency(
                        code_element=self.current_code_element,
                        depends_on_code_element=dep_element,
                        dependency_type='function_call',
                        line=node.lineno,
                        column=node.col_offset
                    )
                    self.dependency_handler(dependency)
        self.generic_visit(node)

    def visit_Import(self, node):
        if 'import' in self.dependency_types:
            for alias in node.names:
                name = alias.asname or alias.name.split('.')[0]
                dep_elements = self.name_to_elements.get(name, set())
                for dep_element in dep_elements:
                    for code_element in self.code_elements_in_file.values():
                        dependency = Dependency(
                            code_element=code_element,
                            depends_on_code_element=dep_element,
                            dependency_type='import',
                            line=node.lineno,
                            column=node.col_offset
                        )
                        self.dependency_handler(dependency)
        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        if 'import_from' in self.dependency_types:
            for alias in node.names:
                name = alias.asname or alias.name
                dep_elements = self.name_to_elements.get(name, set())
                for dep_element in dep_elements:
                    for code_element in self.code_elements_in_file.values():
                        dependency = Dependency(
                            code_element=code_element,
                            depends_on_code_element=dep_element,
                            dependency_type='import_from',
                            line=node.lineno,
                            column=node.col_offset
                        )
                        self.dependency_handler(dependency)
        self.generic_visit(node)

    def visit_Name(self, node):
        if 'name_load' in self.dependency_types and self.current_code_element:
            if isinstance(node.ctx, ast.Load):
                name = node.id
                dep_elements = self.name_to_elements.get(name, set())
                for dep_element in dep_elements:
                    dependency = Dependency(
                        code_element=self.current_code_element,
                        depends_on_code_element=dep_element,
                        dependency_type='name_load',
                        line=node.lineno,
                        column=node.col_offset
                    )
                    self.dependency_handler(dependency)
        self.generic_visit(node)

    def _process_decorators(self, node):
        for decorator in node.decorator_list:
            decorator_name = self._get_full_name(decorator)
            dep_elements = self.name_to_elements.get(decorator_name, set())
            for dep_element in dep_elements:
                dependency = Dependency(
                    code_element=self.current_code_element,
                    depends_on_code_element=dep_element,
                    dependency_type='decorator',
                    line=decorator.lineno,
                    column=decorator.col_offset
                )
                self.dependency_handler(dependency)

    def _process_annotation(self, annotation):
        annotation_name = self._get_full_name(annotation)
        dep_elements = self.name_to_elements.get(annotation_name, set())
        for dep_element in dep_elements:
            dependency = Dependency(
                code_element=self.current_code_element,
                depends_on_code_element=dep_element,
                dependency_type='type_annotation',
                line=getattr(annotation, 'lineno', 0),
                column=getattr(annotation, 'col_offset', 0)
            )
            self.dependency_handler(dependency)

    def _get_definition_full_name(self, node):
        parts = [node.name]
        parent = getattr(node, 'parent', None)
        while parent is not None:
            if isinstance(parent, (ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)):
                parts.append(parent.name)
            parent = getattr(parent, 'parent', None)
        return ".".join(reversed(parts))

    def _get_full_name(self, node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Attribute):
            value = self._get_full_name(node.value)
            if value:
                return f"{value}.{node.attr}"
            else:
                return node.attr
        elif isinstance(node, ast.Call):
            return self._get_full_name(node.func)
        elif isinstance(node, ast.Subscript):
            return self._get_full_name(node.value)
        elif isinstance(node, ast.Index):
            return self._get_full_name(node.value)
        elif isinstance(node, ast.Constant):
            return str(node.value)
        else:
            return None



import argparse
import logging
import sys

from deply import __version__
from deply.deply_runner import DeplyRunner


def main():
    parser = argparse.ArgumentParser(prog="deply", description='Deply - An architecture analysis tool')
    parser.add_argument('-V', '--version', action='store_true', help='Show the version number and exit')
    parser.add_argument('-v', '--verbose', action='count', default=1, help='Increase output verbosity')
    subparsers = parser.add_subparsers(dest='command', help='Sub-commands')
    parser_analyze = subparsers.add_parser('analyze', help='Analyze the project')
    parser_analyze.add_argument('--config', type=str, default="deply.yaml", help="Path to the configuration YAML file")
    parser_analyze.add_argument(
        '--report-format',
        type=str,
        choices=["text", "json", "github-actions"],
        default="text",
        help="Format of the output report"
    )
    parser_analyze.add_argument('--output', type=str, help="Output file for the report")
    parser_analyze.add_argument('--mermaid', action='store_true',
                                help="Generate a Mermaid diagram for layer dependencies (red = violation)")
    parser_analyze.add_argument(
        '--max-violations',
        type=int,
        default=0,
        help="Maximum number of allowed violations before failing"
    )

    parser_analyze.add_argument(
        '--parallel',
        type=int,
        nargs='?',
        default=None,
        const=0,
        help="Enable parallel processing of code elements. "
             "Optionally specify the number of processes to use. "
             "If no number is provided, all available CPU cores will be used."
    )

    args = parser.parse_args()

    if args.version:
        print(f"deply {__version__}")
        sys.exit(0)

    if args.command is None:
        args = parser.parse_args(['analyze'] + sys.argv[1:])

    log_level = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG

    logging.basicConfig(
        level=log_level,
        format='[%(asctime)s] [%(levelname)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    logging.info("Starting Deply analysis...")
    runner = DeplyRunner(args)
    if runner.run():
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    main()
